from __future__ import absolute_import, division, print_function, unicode_literals
from warnings import warn
from iminuit.iminuit_warnings import InitialParamWarning
from iminuit import util as mutil
import numpy as np


def pedantic(self, parameters, kwds, errordef):
    def w(msg):
        warn(msg, InitialParamWarning, stacklevel=3)

    for vn in parameters:
        if vn not in kwds:
            w("Parameter %s does not have initial value. Assume 0." % vn)
        if "error_" + vn not in kwds and "fix_" + mutil.param_name(vn) not in kwds:
            w(
                "Parameter %s is floating but does not have initial step size. Assume 1."
                % vn
            )
    for vlim in mutil.extract_limit(kwds):
        if mutil.param_name(vlim) not in parameters:
            w(
                "%s is given. But there is no parameter %s. Ignore."
                % (vlim, mutil.param_name(vlim))
            )
    for vfix in mutil.extract_fix(kwds):
        if mutil.param_name(vfix) not in parameters:
            w(
                "%s is given. But there is no parameter %s. Ignore."
                % (vfix, mutil.param_name(vfix))
            )
    for verr in mutil.extract_error(kwds):
        if mutil.param_name(verr) not in parameters:
            w(
                "%s float. But there is no parameter %s. Ignore."
                % (verr, mutil.param_name(verr))
            )
    if errordef is None:
        w("errordef is not given. Default to 1.")


def profile(self, vname, bins, bound, args, subtract_min):
    # center value
    val = np.linspace(bound[0], bound[1], bins, dtype=np.double)
    result = np.empty(bins, dtype=np.double)
    pos = self.var2pos[vname]
    n = val.shape[0]
    arg = list(self.args if args is None else args)
    if self.use_array_call:
        varg = np.array(arg, dtype=np.double)
        for i in range(n):
            varg[pos] = val[i]
            result[i] = self.fcn(varg)
    else:
        for i in range(n):
            arg[pos] = val[i]
            result[i] = self.fcn(*arg)
    if subtract_min:
        result -= self.fval
    return val, result


def draw_profile(self, vname, x, y, s, band, text):
    from matplotlib import pyplot as plt

    if s is not None:
        s = np.array(s, dtype=bool)
        x = x[s]
        y = y[s]

    plt.plot(x, y)
    plt.grid(True)
    plt.xlabel(vname)
    plt.ylabel("FCN")

    if vname in self.values:
        v = self.values[vname]
    else:
        v = np.argmin(y)
    vmin = None
    vmax = None
    if (vname, 1) in self.merrors:
        vmin = v + self.merrors[(vname, -1)]
        vmax = v + self.merrors[(vname, 1)]
    if vname in self.errors:
        vmin = v - self.errors[vname]
        vmax = v + self.errors[vname]

    plt.axvline(v, color="r")

    if vmin is not None and band:
        plt.axvspan(vmin, vmax, facecolor="g", alpha=0.5)

    if text:
        plt.title(
            ("%s = %.3g" % (vname, v))
            if vmin is None
            else ("%s = %.3g - %.3g + %.3g" % (vname, v, v - vmin, vmax - v)),
            fontsize="large",
        )

    return x, y


def draw_contour(self, x, y, bins=20, bound=2, args=None, show_sigma=False):
    from matplotlib import pyplot as plt

    vx, vy, vz = self.contour(x, y, bins, bound, args, subtract_min=True)

    v = [self.errordef * ((i + 1) ** 2) for i in range(bound)]

    CS = plt.contour(vx, vy, vz, v, colors=["b", "k", "r"])
    if not show_sigma:
        plt.clabel(CS, v)
    else:
        tmp = dict((vv, r"%i $\sigma$" % (i + 1)) for i, vv in enumerate(v))
        plt.clabel(CS, v, fmt=tmp, fontsize=16)
    plt.xlabel(x)
    plt.ylabel(y)
    plt.axhline(self.values[y], color="k", ls="--")
    plt.axvline(self.values[x], color="k", ls="--")
    plt.grid(True)
    return vx, vy, vz


def draw_mncontour(self, x, y, nsigma=2, numpoints=20):
    from matplotlib import pyplot as plt
    from matplotlib.contour import ContourSet

    c_val = []
    c_pts = []
    for sigma in range(1, nsigma + 1):
        pts = self.mncontour(x, y, numpoints, sigma)[2]
        # close curve
        pts.append(pts[0])
        c_val.append(sigma)
        c_pts.append([pts])  # level can have more than one contour in mpl
    cs = ContourSet(plt.gca(), c_val, c_pts)
    plt.clabel(cs, inline=1, fontsize=10)
    plt.xlabel(x)
    plt.ylabel(y)
    return cs



from __future__ import absolute_import, division, unicode_literals
from iminuit.color import Gradient
from iminuit.repr_text import format_numbers

good_style = "background-color:#92CCA6;"
bad_style = "background-color:#FF7878;"
warn_style = "background-color:#FFF79A;"
backgrounds = ("background-color:#F4F4F4;", "background-color:#FFFFFF;")


class Html:
    def __init__(self):
        self.lines = []

    def __iadd__(self, s):
        self.lines.append(s)
        return self

    def __str__(self):
        s = ""
        for x in self.lines:
            s += x + "\n"
        return s

    def __repr__(self):
        return repr(self.lines)


class Tag:
    def __init__(self, name, html, **kwargs):
        self.name = name
        self.html = html
        # sort keys so that order is same on all platforms
        self.html += (
            " ".join(
                ("<%s" % name,)
                + tuple('%s="%s"' % (k, kwargs[k]) for k in sorted(kwargs))
            )
            + ">"
        )

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.html += "</" + self.name + ">"


def table(html, **kwargs):
    return Tag("table", html, **kwargs)


def tr(html, **kwargs):
    return Tag("tr", html, **kwargs)


def th(html, **kwargs):
    return Tag("th", html, **kwargs)


def td(html, **kwargs):
    return Tag("td", html, **kwargs)


def good(x, should_be):
    return good_style if x == should_be else bad_style


def fmin(sfmin):
    """Display FunctionMinum in html representation"""
    goaledm = 1e-4 * sfmin.tolerance * sfmin.up
    s = Html()
    with table(s):
        with tr(s):
            with td(s, colspan="2", title="Minimum value of function"):
                s += "FCN = %.4G" % sfmin.fval
            with td(
                s,
                colspan="3",
                align="center",
                title="No. of calls in last algorithm and total number of calls",
            ):
                s += "Ncalls = %i (%i total)" % (sfmin.nfcn, sfmin.ncalls)
        with tr(s):
            with td(
                s,
                colspan="2",
                title="Estimated distance to minimum and target threshold",
            ):
                s += "EDM = %.3G (Goal: %G)" % (sfmin.edm, goaledm)
            with td(
                s,
                colspan="3",
                align="center",
                title="Increase in FCN which corresponds to 1 standard deviation",
            ):
                s += "up = %.1f" % sfmin.up
        with tr(s):
            with td(s, align="center", title="Validity of the migrad call"):
                s += "Valid Min."
            with td(s, align="center", title="Validity of parameters"):
                s += "Valid Param."
            with td(s, align="center", title="Is EDM above goal EDM?"):
                s += "Above EDM"
            with td(
                s,
                colspan="2",
                align="center",
                title="Did last migrad call reach max call limit?",
            ):
                s += "Reached call limit"
        with tr(s):
            with td(s, align="center", style=good(sfmin.is_valid, True)):
                s += "%s" % sfmin.is_valid
            with td(s, align="center", style=good(sfmin.has_valid_parameters, True)):
                s += "%s" % sfmin.has_valid_parameters
            with td(s, align="center", style=good(sfmin.is_above_max_edm, False)):
                s += "%s" % sfmin.is_above_max_edm
            with td(
                s,
                colspan="2",
                align="center",
                style=good(sfmin.has_reached_call_limit, False),
            ):
                s += "%s" % sfmin.has_reached_call_limit
        with tr(s):
            with td(s, align="center", title="Did Hesse fail?"):
                s += "Hesse failed"
            with td(s, align="center", title="Has covariance matrix"):
                s += "Has cov."
            with td(s, align="center", title="Is covariance matrix accurate?"):
                s += "Accurate"
            with td(s, align="center", title="Is covariance matrix positive definite?"):
                s += "Pos. def."
            with td(
                s, align="center", title="Was positive definiteness enforced by Minuit?"
            ):
                s += "Forced"
        with tr(s):
            with td(s, align="center", style=good(sfmin.hesse_failed, False)):
                s += "%s" % sfmin.hesse_failed
            with td(s, align="center", style=good(sfmin.has_covariance, True)):
                s += "%s" % sfmin.has_covariance
            with td(s, align="center", style=good(sfmin.has_accurate_covar, True)):
                s += "%s" % sfmin.has_accurate_covar
            with td(s, align="center", style=good(sfmin.has_posdef_covar, True)):
                s += "%s" % sfmin.has_posdef_covar
            with td(s, align="center", style=good(sfmin.has_made_posdef_covar, False)):
                s += "%s" % sfmin.has_made_posdef_covar
    return str(s)


def merror(me):
    mel, meu = format_numbers(me.lower, me.upper)
    s = Html()
    with table(s):
        with tr(s):
            with th(s, title="Parameter name"):
                s += me.name
            with td(s, colspan="2", style=good(me.is_valid, True), align="center"):
                s += "Valid" if me.is_valid else "Invalid"
        with tr(s):
            with td(s, title="Lower and upper minos error of the parameter"):
                s += "Error"
            with td(s):
                s += mel
            with td(s):
                s += meu
        with tr(s):
            with td(s, title="Validity of lower/upper minos error"):
                s += "Valid"
            with td(s, style=good(me.lower_valid, True)):
                s += "%s" % me.lower_valid
            with td(s, style=good(me.upper_valid, True)):
                s += "%s" % me.upper_valid
        with tr(s):
            with td(s, title="Did scan hit limit of any parameter?"):
                s += "At Limit"
            with td(s, style=good(me.at_lower_limit, False)):
                s += "%s" % me.at_lower_limit
            with td(s, style=good(me.at_upper_limit, False)):
                s += "%s" % me.at_upper_limit
        with tr(s):
            with td(s, title="Did scan hit function call limit?"):
                s += "Max FCN"
            with td(s, style=good(me.at_lower_max_fcn, False)):
                s += "%s" % me.at_lower_max_fcn
            with td(s, style=good(me.at_upper_max_fcn, False)):
                s += "%s" % me.at_upper_max_fcn
        with tr(s):
            with td(s, title="New minimum found when doing scan?"):
                s += "New Min"
            with td(s, style=good(me.lower_new_min, False)):
                s += "%s" % me.lower_new_min
            with td(s, style=good(me.upper_new_min, False)):
                s += "%s" % me.upper_new_min
    return str(s)


def params(mps):
    s = Html()
    with table(s):
        # header
        with tr(s, style=backgrounds[0]):
            s += "<td/>"
            with th(s, title="Variable name"):
                s += "Name"
            with th(s, title="Value of parameter"):
                s += "Value"
            with th(s, title="Hesse error"):
                s += "Hesse Error"
            with th(s, title="Minos lower error"):
                s += "Minos Error-"
            with th(s, title="Minos upper error"):
                s += "Minos Error+"
            with th(s, title="Lower limit of the parameter"):
                s += "Limit-"
            with th(s, title="Upper limit of the parameter"):
                s += "Limit+"
            with th(s, title="Is the parameter fixed in the fit"):
                s += "Fixed"

        mes = mps.merrors

        # body
        for i, mp in enumerate(mps):
            if mes and mp.name in mes:
                me = mes[mp.name]
                v, e, mem, mep = format_numbers(mp.value, mp.error, me.lower, me.upper)
            else:
                e, v = format_numbers(mp.error, mp.value)
                mem = ""
                mep = ""
            with tr(s, style=backgrounds[(i + 1) % 2]):
                with td(s):
                    s += str(i)
                with td(s):
                    s += mp.name
                with td(s):
                    s += v
                with td(s):
                    s += e
                with td(s):
                    s += mem
                with td(s):
                    s += mep
                with td(s):
                    s += "%.3G" % mp.lower_limit if mp.lower_limit is not None else ""
                with td(s):
                    s += "%.3G" % mp.upper_limit if mp.upper_limit is not None else ""
                with td(s):
                    s += "yes" if mp.is_fixed else ("CONST" if mp.is_const else "")
    return str(s)


def matrix(m):
    is_correlation = True
    for i in range(len(m)):
        if m[i][i] != 1.0:
            is_correlation = False
            break

    if not is_correlation:
        n = len(m)
        args = []
        for mi in m:
            for mj in mi:
                args.append(mj)
        nums = format_numbers(*args)

    grad = Gradient(
        (-1.0, 120.0, 120.0, 250.0),
        (0.0, 250.0, 250.0, 250.0),
        (1.0, 250.0, 100.0, 100.0),
    )

    s = Html()
    with table(s):
        with tr(s):
            s += "<td/>\n"
            for v in m.names:
                with th(s):
                    s += v
        for i, v in enumerate(m.names):
            with tr(s):
                with th(s):
                    s += v
                for j in range(len(m.names)):
                    val = m[i][j]
                    if is_correlation:
                        if i == j:
                            with td(s):
                                s += " 1.00"
                        else:
                            color = grad.rgb(val)
                            with td(s, style="background-color:" + color):
                                s += "%5.2f" % val
                    else:
                        if i == j:
                            with td(s):
                                s += nums[n * i + j]
                        else:
                            color = grad.rgb(val / (m[i][i] * m[j][j]) ** 0.5)
                            with td(s, style="background-color:" + color):
                                s += nums[n * i + j]

    return str(s)



from __future__ import absolute_import, division, print_function, unicode_literals
from math import sqrt, exp, cos, pi, e
from iminuit import Minuit
from iminuit.tests.utils import assert_allclose


def rosenbrock(x, y):
    """Rosenbrock function, minimum at (1,1) with value of 0"""
    return (1 - x) ** 2 + 100 * (y - x ** 2) ** 2


def test_rosenbrock():
    m = Minuit(rosenbrock, x=0, y=0, pedantic=False, print_level=1)
    m.tol = 1e-4
    m.migrad()
    assert_allclose(m.fval, 0, atol=1e-6)
    assert_allclose(m.values["x"], 1.0, atol=1e-3)
    assert_allclose(m.values["y"], 1.0, atol=1e-3)


def ackleys(x, y):
    term1 = -20 * exp(-0.2 * sqrt(0.5 * (x ** 2 + y ** 2)))
    term2 = -exp(0.5 * (cos(2 * pi * x) + cos(2 * pi * y)))
    return term1 + term2 + 20 + e


def test_ackleys():
    m = Minuit(
        ackleys, x=0.3, y=-0.2, error_x=1.7, error_y=1.7, pedantic=False, print_level=0
    )
    m.tol = 1e-4
    m.migrad()

    assert_allclose(m.fval, 0, atol=1e-6)
    assert_allclose(m.args, [0, 0], atol=1e-6)


def beale(x, y):
    term1 = 1.5 - x + x * y
    term2 = 2.25 - x + x * y ** 2
    term3 = 2.625 - x + x * y ** 3

    return term1 * term1 + term2 * term2 + term3 * term3


def test_beale():
    m = Minuit(beale, x=0.5, y=0.25, pedantic=False, print_level=0)
    m.tol = 1e-4
    m.migrad()
    assert_allclose(m.fval, 0, atol=1e-6)
    assert_allclose(m.args, [3, 0.5], atol=1e-3)
    assert m.fval < 1e-6


def matyas(x, y):
    return 0.26 * (x ** 2 + y ** 2) - 0.48 * x * y


def test_matyas():
    m = Minuit(matyas, x=0.5, y=0.5, pedantic=False, print_level=0)
    m.tol = 1e-4
    m.migrad()

    assert m.fval < 1e-26
    assert_allclose(m.args, [0, 0], atol=1e-12)


def test_matyas_oneside():
    """One-sided limit when the minimum is in the forbidden region"""
    m = Minuit(
        matyas, x=2 + 0.5, y=0.5, limit_x=(1, None), pedantic=False, print_level=0
    )
    m.tol = 1e-4
    m.migrad()
    assert_allclose(m.args, [1, 0.923], atol=1e-3)



from __future__ import absolute_import, division, print_function
from iminuit.util import (
    fitarg_rename,
    true_param,
    param_name,
    extract_iv,
    extract_limit,
    extract_error,
    extract_fix,
    remove_var,
    arguments_from_docstring,
    Matrix,
    FMin,
    Param,
    MError,
    Params,
    MigradResult,
)
import pytest


def test_fitarg_rename():
    fitarg = {"x": 1, "limit_x": (2, 3), "fix_x": True, "error_x": 10}

    def ren(x):
        return "z_" + x

    newfa = fitarg_rename(fitarg, ren)
    assert "z_x" in newfa
    assert "limit_z_x" in newfa
    assert "error_z_x" in newfa
    assert "fix_z_x" in newfa
    assert len(newfa) == 4


def test_fitarg_rename_strprefix():
    fitarg = {"x": 1, "limit_x": (2, 3), "fix_x": True, "error_x": 10}
    newfa = fitarg_rename(fitarg, "z")
    assert "z_x" in newfa
    assert "limit_z_x" in newfa
    assert "error_z_x" in newfa
    assert "fix_z_x" in newfa
    assert len(newfa) == 4


def test_true_param():
    assert true_param("N") is True
    assert true_param("limit_N") is False
    assert true_param("error_N") is False
    assert true_param("fix_N") is False


def test_param_name():
    assert param_name("N") == "N"
    assert param_name("limit_N") == "N"
    assert param_name("error_N") == "N"
    assert param_name("fix_N") == "N"


def test_extract_iv():
    d = dict(k=1.0, limit_k=1.0, error_k=1.0, fix_k=1.0)
    ret = extract_iv(d)
    assert "k" in ret
    assert "limit_k" not in ret
    assert "error_k" not in ret
    assert "fix_k" not in ret


def test_extract_limit():
    d = dict(k=1.0, limit_k=1.0, error_k=1.0, fix_k=1.0)
    ret = extract_limit(d)
    assert "k" not in ret
    assert "limit_k" in ret
    assert "error_k" not in ret
    assert "fix_k" not in ret


def test_extract_error():
    d = dict(k=1.0, limit_k=1.0, error_k=1.0, fix_k=1.0)
    ret = extract_error(d)
    assert "k" not in ret
    assert "limit_k" not in ret
    assert "error_k" in ret
    assert "fix_k" not in ret


def test_extract_fix():
    d = dict(k=1.0, limit_k=1.0, error_k=1.0, fix_k=1.0)
    ret = extract_fix(d)
    assert "k" not in ret
    assert "limit_k" not in ret
    assert "error_k" not in ret
    assert "fix_k" in ret


def test_remove_var():
    dk = dict(k=1, limit_k=1, error_k=1, fix_k=1)
    dl = dict(l=1, limit_l=1, error_l=1, fix_l=1)
    dm = dict(m=1, limit_m=1, error_m=1, fix_m=1)
    dn = dict(n=1, limit_n=1, error_n=1, fix_n=1)
    d = {}
    d.update(dk)
    d.update(dl)
    d.update(dm)
    d.update(dn)

    ret = remove_var(d, ["k", "m"])
    for k in dk:
        assert k not in ret
    for k in dl:
        assert k in ret
    for k in dm:
        assert k not in ret
    for k in dn:
        assert k in ret


def test_arguments_from_docstring():
    s = "f(x, y, z)"
    a = arguments_from_docstring(s)
    assert a == ["x", "y", "z"]
    # this is a hard one
    s = "Minuit.migrad( int ncall_me =10000, [resume=True, int nsplit=1])"
    a = arguments_from_docstring(s)
    assert a == ["ncall_me", "resume", "nsplit"]


def test_Matrix():
    x = Matrix(("a", "b"), [[1, 2], [3, 4]])
    assert x[0] == (1, 2)
    assert x[1] == (3, 4)
    assert x == ((1, 2), (3, 4))
    assert repr(x) == "((1, 2), (3, 4))"
    with pytest.raises(TypeError):
        x[0] = (1, 2)
    with pytest.raises(TypeError):
        x[0][0] = 1


def test_Param():
    keys = "number name value error is_const is_fixed has_limits has_lower_limit has_upper_limit lower_limit upper_limit".split()
    values = 3, "foo", 1.2, 3.4, False, False, True, True, False, 42, None
    p = Param(*values)

    assert p.has_lower_limit == True
    assert p.has_upper_limit == False
    assert p["has_lower_limit"] == True
    assert p["lower_limit"] == 42
    assert p["upper_limit"] == None
    assert "upper_limit" in p
    assert "foo" not in p

    assert [key for key in p] == keys
    assert p.keys() == tuple(keys)
    assert p.values() == values
    assert p.items() == tuple((k, v) for k, v in zip(keys, values))

    assert (
        str(p)
        == "Param(number=3, name='foo', value=1.2, error=3.4, is_const=False, is_fixed=False, has_limits=True, has_lower_limit=True, has_upper_limit=False, lower_limit=42, upper_limit=None)"
    )


def test_MError():
    keys = "name is_valid lower upper lower_valid upper_valid at_lower_limit at_upper_limit at_lower_max_fcn at_upper_max_fcn lower_new_min upper_new_min nfcn min".split()
    values = (
        "Foo",
        True,
        0.1,
        1.2,
        True,
        True,
        False,
        False,
        False,
        False,
        0.1,
        1.2,
        42,
        0.2,
    )

    m = MError(*values)

    assert m.keys() == tuple(keys)
    assert m.values() == values
    assert m.name == "Foo"
    assert m["name"] == "Foo"


def test_FMin():
    keys = "fval edm tolerance nfcn ncalls up is_valid has_valid_parameters has_accurate_covar has_posdef_covar has_made_posdef_covar hesse_failed has_covariance is_above_max_edm has_reached_call_limit".split()
    values = (
        0.2,
        1e-3,
        0.1,
        10,
        10,
        1.2,
        False,
        False,
        False,
        False,
        False,
        False,
        False,
        False,
        False,
    )

    f = FMin(*values)

    assert f.keys() == tuple(keys)
    assert f.values() == values
    assert f.fval == 0.2
    assert f["fval"] == 0.2


def test_MigradResult():
    fmin = FMin(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14)
    params = Params([], None)
    mr = MigradResult(fmin, params)
    assert mr.fmin is fmin
    assert mr[0] is fmin
    assert mr.params is params
    assert mr[1] is params
    a, b = mr
    assert a is fmin
    assert b is params
